/************************************************************************
 *
 * Author:     Luis Gerardo de la Fraga     fraga@cs.cinvestav.mx
 *
 * Computer Science Deparment, Cinvestav.
 *
 * Copyright (c) 2007, Cinvestav.
 *
 * Permission is granted to copy and distribute this file, for noncommercial
 * use, provided (a) this copyright notice is preserved, and (b) no attempt
 * is made to restrict redistribution of this file,

 *  All comments concerning this program may be sent to the
 *  e-mail address 'fraga@cs.cinvestav.mx'
 *
 **************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <string.h>

/**
#define DEBUG 0
   Activating this DEBUG flag produce:
   The best guy is printed every 20 generations
   And this flag can be activated at compiled time as:

   gcc -O2 -DDEBUG -o edellipse edellipse.c -lm
**/
#define NPARAM  6

#define A  0
#define B  1
#define XC 2
#define YC 3
#define ALPHA 4
#define FIT   5

typedef struct param {
	double p[NPARAM];
} PARAM;

typedef struct point {
	double x, y;
} POINT;

typedef struct data {
	int n;
	POINT *pts;
	POINT limits[5]; /** For a, b, xc, yc, and alpha **/
	int ngen, popsize;
	PARAM *guys0;
	PARAM *guys1;
	double Pc, Pf; 
} DATA;

float ran0( int * idum )
{
	static float y,maxran,v[98];
	unsigned int semilla;
	static int iff=0;
	int j;

	if ( *idum <0 || iff == 0) {
		semilla = (unsigned int) time ( NULL );
		srand( semilla );
		iff=1;
		*idum=1;
		maxran = RAND_MAX + 1.0;
		for (j=1;j<=97;j++) v[j]=rand();
		y=rand();
	}
	j= 1 + (int)( 97.0f*y/maxran );
/** if (j > 97 || j < 1) printf("RAN0: This cannot happen.\n"); **/
	y=v[j];
	v[j]=rand();
	return (y/maxran);
}

double find_point_in_hyperbola( double a, double b, double x0, double y0 )
{
	double l0, l1, v, v1, v2;
	double f, df, s, c;
	int i;

	l0 = asinh( y0/b );

	v = a*a + b*b;
	for ( i=0; i<=3; i++ ) {
		s = sinh(l0);	
		c = cosh(l0);	
		v1 = y0*b;
		v2 = x0*a;
		f = v1*c + v2*s - v*c*s;
		df = v1*s + v2*c - v*(c*c+s*s);

		l1 = l0 - f/df;
		if ( fabs(l1-l0)< 0.00001 ){
			return( l1 );
		}

		l0 = l1;
	}
	return l0;
}


void evaluateGuy( PARAM *ptrg, POINT *ptrp, int n ) 
{
	int i;
	double c, s, d, m;
	POINT ptran, prota, dif;

#ifdef CONSTR
	ptrg->p[B] = ptrg->p[A]/20.0;
#endif
/**
	if ( ptrg->p[A] < ptrg->p[B] ) {
		s = ptrg->p[A];
		ptrg->p[A] = ptrg->p[B];
		ptrg->p[B] = s;
	}	
**/

	c = cos( ptrg->p[ALPHA] );
	s = sin( ptrg->p[ALPHA] );

	d = 0;
	for ( i=0; i<n; i++ ) {
		ptran.x = ptrp[i].x - ptrg->p[XC];  
		ptran.y = ptrp[i].y - ptrg->p[YC];  

		prota.x =  ptran.x*c + ptran.y*s;
		prota.y = -ptran.x*s + ptran.y*c;

		if( prota.x < 0.0 ) prota.x  = -prota.x;

		m = find_point_in_hyperbola( ptrg->p[A], ptrg->p[B], prota.x, prota.y ); 

		dif.x = ptrg->p[A]*cosh(m) - prota.x;
		dif.y = ptrg->p[B]*sinh(m) - prota.y;

		d += sqrt( dif.x*dif.x + dif.y*dif.y );

		/** d += dif.x*dif.x + dif.y*dif.y;  **/
	}
	/** fprintf( stderr, "d = %lf\n", d ); **/
	

	ptrg->p[FIT] = d;
}

/** The main algorithm **/	
void eDifferential( DATA *pd )
{
	int i, i0, i1, i2;
	int n, seed=-6;
	int j, k, jrand;

	/** Evaluate to all guys **/
	for(i=0; i<pd->popsize; i++)
		evaluateGuy( &pd->guys0[i], pd->pts, pd->n ); 

	n = pd->popsize;
	for(i=0; i<pd->ngen; i++){
		for(k=0; k<pd->popsize; k++){
			/** We select three different parents **/
			i0 = (int)(ran0( &seed )*n);
			do { 
				i1 = (int)(ran0( &seed )*n);
			} while( i1 == i0 );
			do { 
				i2 = (int)(ran0( &seed )*n);
			} while( i2==i1 || i2 == i0 );

			/** We apply the crossoverBin **/
#ifndef CONSTR
			jrand = (int)(5.0*ran0( &seed ));
#else
			jrand = (int)(3.0*ran0( &seed ));
			if ( jrand >= B ) jrand++;
#endif

#ifndef CONSTR
			for ( j=0; j<5; j++ ) {
#else
			for ( j=0; j<4; j++ ) {
				/** An example of how to use constraints b=a/20, alpha=pi/2 **/
				if ( j == B ) j++;
#endif
				if ( (ran0( &seed ) < pd->Pc) || j==jrand ) {
					pd->guys1[k].p[j] = pd->guys0[i2].p[j] +
					    pd->Pf * (pd->guys0[i0].p[j] - pd->guys0[i1].p[j]); 

				/** We avoid this value be outside its bounds **/
				if ( pd->guys1[k].p[j] < pd->limits[j].x ||
				     pd->guys1[k].p[j] > pd->limits[j].y ) {
					pd->guys1[k].p[j] = (double)(((double)(
					   (long)(fabs(pd->guys0[k].p[j]*1e8))%
					         (long)((pd->limits[j].y-pd->limits[j].x)*1e8))/1e8)
					        + pd->limits[j].x);
					}
				}
				else{
					pd->guys1[k].p[j] = pd->guys0[k].p[j];
				}
			}
			/** Evaluation of the new guy 
			fprintf( stderr, "%lf %lf %lf %lf %lf\n",
			     pd->guys1[k].p[A], pd->guys1[k].p[B], pd->guys1[k].p[XC],
			     pd->guys1[k].p[YC], pd->guys1[k].p[ALPHA] );
			**/
			evaluateGuy( &pd->guys1[k], pd->pts, pd->n ); 

			/** Copy the new guy if it is better **/
			if( pd->guys1[k].p[FIT] < pd->guys0[k].p[FIT] ){
				for ( j=0; j<6; j++ ) 
					pd->guys0[k].p[j] = pd->guys1[k].p[j];
			}
		} /** End of one generation **/
#ifdef DEBUG 
		/** We find the best guy **/
		if ( !(i%20) ) { 
			i0 = 0;
			for ( j=1; j<n; j++ ) {
				if ( pd->guys0[j].p[FIT] < pd->guys0[i0].p[FIT] )
					i0 = j;
			}
			printf( "%d %lf %lf %lf %lf %lf %lf\n", i,
				pd->guys0[i0].p[A], pd->guys0[i0].p[B], 
				pd->guys0[i0].p[XC], pd->guys0[i0].p[YC], 
				pd->guys0[i0].p[ALPHA], pd->guys0[i0].p[FIT]  ); 
		}
#endif
	}
	/** We find the best guy **/
	i0 = 0;
	for ( j=1; j<n; j++ ) {
		if ( pd->guys0[j].p[FIT] < pd->guys0[i0].p[FIT] )
			i0 = j;
	}
	printf( "%d %lf %lf %lf %lf %lf %lf\n", i,
		pd->guys0[i0].p[A], pd->guys0[i0].p[B], 
		pd->guys0[i0].p[XC], pd->guys0[i0].p[YC], 
		pd->guys0[i0].p[ALPHA], pd->guys0[i0].p[FIT]  ); 
}



int initialization( DATA *pd )
{
	int i, j, seed=-3; 
	double range[5];

	pd->pts = (POINT *)malloc( pd->n * sizeof(POINT) );
	if ( pd->pts == NULL ) return 1;

	pd->guys0 = (PARAM *)malloc( pd->popsize * sizeof(PARAM) );
	if ( pd->guys0 == NULL ){
		free( pd->pts );
		return 1;
	}
	
	pd->guys1 = (PARAM *)malloc( pd->popsize * sizeof(PARAM) );
	if ( pd->guys1 == NULL ){
		free( pd->guys0 );
		free( pd->pts );
		return 1;
	}

	/** And we initializate the guys with random values **/
	for( j=0; j<5; j++ ) 
		range[j] = pd->limits[j].y - pd->limits[j].x;
		
	for( i=0; i<pd->popsize; i++ ) {
		for( j=0; j<5; j++ ) 
			pd->guys0[i].p[j] = pd->limits[j].x + range[j]*ran0(&seed);
	}

	return 0;
}

int read_input_data( char *namefile, DATA *pd )
{
	FILE *fp;
	double x, y;
	int i, n;
	char line[64];

	if ( (fp=fopen( namefile, "r" ))==NULL ){
		fprintf( stderr, "Error: can't open %s file\n", namefile );
		return 1;
	}	
	i=0;
	while ( fgets( line, 64, fp ) != NULL ) {
		n = sscanf( line, "%lf %lf", &x, &y );

		if ( n == 2 ){
			pd->pts[i].x = x;
			pd->pts[i].y = y;
			i++;
		} 

	}
	fclose( fp );
	/**
	for (i=0; i<pd->n; i++ ) 
		printf ("%d %lf %lf\n", i, pd->pts[i].x, pd->pts[i].y );
	pd->guys0[0].p[A] = 6.51872; 
	pd->guys0[0].p[B] = 3.03189;
	pd->guys0[0].p[XC] = 2.69962;
	pd->guys0[0].p[YC] = 3.81596;
	pd->guys0[0].p[ALPHA] = 0.35962;

	evaluateGuy( &pd->guys0[0], pd->pts, pd->n ); 
	printf( "Error = %lf\n", pd->guys0[0].p[FIT] );
**/
	 
	if ( i != pd->n ) {
		fprintf( stderr, "Error: I read %d points\n", i );
		return 1;
	}
	return 0;
}

void free_memory ( DATA *pd )
{
	free( pd->guys0 );
	free( pd->guys1 );
	free( pd->pts );
}

int main( int argc, char * argv[] ) 
{
	DATA d;

	char file[64];

	if( argc != 17 ) { 
		fprintf( stderr, "Args: a_min a_max b_min b_max xc_min xc_max\n\tyc_min yc_max alpha_min alpha_max n data_file ngen popsize Pc Pf\n" );
		return 1;
	}

	d.limits[0].x = strtod( argv[1], (char **)NULL);
	d.limits[0].y = strtod( argv[2], (char **)NULL);

	d.limits[1].x = strtod( argv[3], (char **)NULL);
	d.limits[1].y = strtod( argv[4], (char **)NULL);

	d.limits[2].x = strtod( argv[5], (char **)NULL);
	d.limits[2].y = strtod( argv[6], (char **)NULL);

	d.limits[3].x = strtod( argv[7], (char **)NULL);
	d.limits[3].y = strtod( argv[8], (char **)NULL);

	d.limits[4].x = strtod( argv[9], (char **)NULL);
	d.limits[4].y = strtod( argv[10], (char **)NULL);

	d.n = (int)strtol( argv[11], (char **)NULL, 10); 
	strncpy( file, argv[12], 64 );

	d.ngen = (int)strtol( argv[13], (char **)NULL, 10); 
	d.popsize = (int)strtol( argv[14], (char **)NULL, 10);
	if ( d.popsize < 3 ){
		fprintf( stderr, "Population size must be greater than 3\n" );
		return 1;
	}
	d.Pc = strtod( argv[15], (char **)NULL);
	d.Pf = strtod( argv[16], (char **)NULL);

	if( initialization( &d ) ) {
		fprintf( stderr, "Error allocating memory\n" );
		return 1;
	}
	if( read_input_data( file, &d ) ) {
		free_memory ( &d );
		fprintf( stderr, "Error reading data\n" );
		return 1;
	}
	/** fprintf( stderr, "Read %d data pairs\n", d.n ); **/

	eDifferential( &d );

	free_memory ( &d );
	return 0;
}
